#include <linux/init.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/wait.h>
#include <linux/ipv6.h>
#include <linux/crypto.h>
#include <linux/scatterlist.h>
#include <linux/time.h>
#include <linux/netfilter.h>
#include <net/ipv6.h>
#include <asm/atomic.h>

#include <linux/netfilter_ipv6.h>

#include "kernel_ver.h"
#include "rmd160.h"
#include "pla_header.h"
#include "pla_netlink.h"
#include "pla_crypto.h"
#include "pla_crypto_impl.h"
#include "pla_policy.h"
#include "pla_src_id.h"
#include "pla_state.h"
#include "pla_conn.h"

#define PLA_PROC_STAT_NAME "net/pla_stats"

atomic_t processing_packets      = ATOMIC_INIT(0);  /* Number of packets under PLA processing  */
atomic_t pla_stats_total_packets = ATOMIC_INIT(0);

struct pla_call_state 
{
	struct sk_buff *skb;
	struct nf_info *info;
	struct pla_hdr *pla_hdr;

	void (*callback) (unsigned long state_ptr);

	struct pla_crypto_task task;
};

int pla_read_stats_proc(char *page, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	
	len += sprintf(page + len, "%u ", atomic_read(&pla_stats_total_packets));
	len += sprintf(page + len, "%d\n", atomic_read(&processing_packets) );
	
	*eof = 1;
	return len;
}

static void pla_immediate_verdict(struct sk_buff *skb, struct nf_info *info, int verdict)
{
	atomic_dec(&processing_packets);
	nf_reinject(skb, info, verdict);

	return;
}

static void pla_immediate_verdict_state(struct pla_call_state *cs, int verdict)
{
	pla_immediate_verdict(cs->skb, cs->info, verdict);

	return;
}

struct pla_call_state* alloc_call_state(struct sk_buff *skb, struct nf_info *info)
{
	struct pla_call_state *cs;

	cs = kmalloc(sizeof(*cs), GFP_ATOMIC);
	if (!cs)
		goto drop_out;
	cs->skb = skb;
	cs->info = info;

	cs->task.type = 0;
	cs->task.code = 0;

	cs->task.data_len = 0; /* The data itself can remain uninitialized */

drop_out:
	return cs;
}

int pla_setup_crypto(struct pla_crypto_task *task, int type)
{
	task->type = type;
	task->data_len = 0;

	return 0;
}

int pla_packet_count_hash(struct sk_buff *skb, char *hash_result)
{
	RMD160_CTX ctx;

	int err = 0;
	struct ipv6_opt_hdr *hbh;
	unsigned char *past_hbh;

	if (skb->nh.ipv6h->nexthdr != NEXTHDR_HOP) {
		err = -1;
		goto out;
	}

	hbh = (struct ipv6_opt_hdr*) (skb->nh.ipv6h + 1);
	past_hbh = ((char*) hbh) + ipv6_optlen(hbh);

	RMD160Init(&ctx);
	RMD160Update(&ctx, past_hbh, ntohs(skb->nh.ipv6h->payload_len) + sizeof(struct ipv6hdr) - (past_hbh - skb->nh.raw));
	RMD160Final(hash_result, &ctx);

out:
	return err;
}

signed int pla_create_verify_request(struct sk_buff *skb, struct pla_hdr *pla_hdr, struct crypto_verify_request *vreq)
{
	int ret = 0;
	
	if (pla_packet_count_hash(skb, vreq->hash) < 0) {
		ret = -1;
		goto out;
	}

#if 0
	switch (pla_state_get(vreq->ttp_pubkey)) {
	case 0:
		ret = 0;
		break;
	case -1:        /* Unknown TTP, ICMP already sent recently */
		ret = -1;
		goto out;

#if 0
	case -2:        /* Unknown TTP, No ICMP sent yet */
		pla_send_icmp(PLA_CERT_REQUEST, PTR_TO_ID_HASH, PACKET_LENGTH);
		ret = -1;
		break;
#endif

	default:
		BUG();
		break;
	}
#else

	
#endif

	/* Always trust the public key in the packet. jlu XXX: This is wrong BTW.... */
	memcpy(vreq->ttp_pubkey, &pla_hdr->public_key, sizeof(vreq->ttp_pubkey));

	memcpy(vreq->signature, &pla_hdr->signature, sizeof(vreq->signature) );
	memcpy(vreq->cert, &pla_hdr->certificate, sizeof(vreq->cert) );
	memcpy(vreq->id_hash, &pla_hdr->id_hash, sizeof(vreq->id_hash));

	ret = sizeof(struct crypto_verify_request);
out:
	return ret;
}

int pla_setup_crypto_call_verify(struct sk_buff *skb, struct nf_info *info, struct pla_hdr *pla_hdr, int task_type, 
				 void (*callback) (void *call_state) )
{
	int err;

	struct pla_call_state *cs = NULL;

	/* Verify that the public key belongs to a valid node */
	// jlu XXX TBD.


	cs = alloc_call_state(skb, info);
	if (!cs) {
		err = -ENOMEM;
		goto drop_out;
	}
	
	cs->pla_hdr = pla_hdr;
	pla_setup_crypto(&cs->task, task_type);
	if ( (cs->task.data_len = pla_create_verify_request(skb, pla_hdr, &cs->task.td.vreq)) < 0) {
		err = -ENOMEM;
		goto drop_out;
	}

	if ( (err = pla_crypto_call(callback, cs, &cs->task)) < 0)
		goto drop_out;

	return 0;
	
drop_out:
	if (cs)
		kfree(cs);

	return err;
}

int pla_setup_crypto_call_sign(struct sk_buff *skb, struct nf_info *info, struct pla_hdr *pla_hdr, 
			       void (*callback) (void *call_state))
{
	struct crypto_sign_request *sr;
	int err = 0;
	struct pla_call_state *cs = NULL;
	struct timeval tv;

	/* Copy time into packet */
	do_gettimeofday(&tv);
	pla_hdr->creation_time = cpu_to_be32(tv.tv_sec);

	if (!(cs = alloc_call_state(skb, info)))
		goto drop_out;

	/* Create crypto request message */
	pla_setup_crypto(&cs->task, CRYPTO_TASK_TYPE_SIGN);
	sr = &cs->task.td.sreq;

	pla_src_key_copy(pla_hdr->public_key, sr->privkey, pla_hdr->certificate, &pla_hdr->seqnum, pla_hdr->id_hash, &pla_hdr->ttp_id);
	memcpy(sr->cert, pla_hdr->certificate, sizeof(sr->cert) );

	pla_packet_count_hash(skb, sr->hash);
	cs->task.data_len = sizeof(struct crypto_sign_request);
	cs->pla_hdr = pla_hdr;

	if ( (err = pla_crypto_call(callback, cs, &cs->task)) < 0)
		goto drop_out;
	
	return 0;

drop_out:
	if (cs)
		kfree(cs);

	return err;
}

void pla_out_path_finish(void *call_state)
{
	struct pla_call_state *state;
	struct pla_hdr *pla_hdr;

	struct crypto_sign_response *sres;

	state = (struct pla_call_state*) call_state;
	pla_hdr = state->pla_hdr;
	sres = &state->task.td.sresp;
	
	memcpy(&pla_hdr->signature, sres->signature, sizeof(pla_hdr->signature) );
	pla_immediate_verdict_state(call_state, state->task.code == CRYPTO_TASK_CODE_OK ? NF_ACCEPT : NF_DROP);
	
	kfree(state);

	return;
}

int pla_out_path(struct sk_buff **skb, struct nf_info *info, struct pla_hdr *pla_hdr)
{
	int err = -1;

	if (!pla_hdr) {
		/* TCP output path for packets associated with a socket. RST packets have no skb->sk set. */
		if ( (*skb)->sk && (*skb)->sk->sk_type == SOCK_STREAM) {
			if (pla_add_pla_option(*skb) < 0)
				goto drop_out;
			if (pla_add_header(skb) < 0) 
				goto drop_out;
		} else {
			err = pla_add_header(skb);
			if (err)
				goto drop_out;
		}

		pla_hdr = pla_find_pla_header(*skb);
	}

	if ( (err = pla_setup_crypto_call_sign(*skb, info, pla_hdr, pla_out_path_finish)) < 0)
		goto drop_out;

	return 0;

drop_out:
	return err;
}

int complete_verify(void *call_state)
{
	struct pla_call_state *state = (struct pla_call_state*) call_state;
	struct pla_hdr *pla_hdr = state->pla_hdr;
	int ret = 0;

	if (state->task.td.vresp.result != 1) {
		ret = -1;
		goto out;
	}

	/* Signature and certificate are ok. Verify time and seqnum. */
#if 1
	if (pla_conn_find_add(pla_hdr->id_hash, be64_to_cpu(pla_hdr->seqnum), be32_to_cpu(pla_hdr->creation_time) ) < 0) {
		ret = -1;
		goto out;
	}
#else
	// jlu XXX: Check sanity...
	if (pla_conn_find_add(pla_hdr->id_hash, be64_to_cpu(pla_hdr->seqnum), be32_to_cpu(pla_hdr->creation_time) ) < 0) {
		state->skb->nfmark = 2;
	}		       
#endif

out:
	return ret;
}

void pla_finish_verify(void *call_state)
{
	pla_immediate_verdict_state(call_state, complete_verify(call_state) >= 0 ? NF_ACCEPT : NF_DROP);
	kfree(call_state);

	return;
}

void pla_finish_verify_strip(void *call_state)
{
	struct pla_call_state *cs = call_state;

	if (complete_verify(call_state) < 0) 
		goto drop_out;
		
	/* Packet has been accepted. Strip the outermost PLA header */
	if (pla_strip_pla_header(&cs->skb) < 0)
		goto drop_out;

	pla_immediate_verdict(cs->skb, cs->info, NF_ACCEPT);

	kfree(call_state);
	return;

drop_out:
	pla_immediate_verdict_state(call_state, NF_DROP);
	kfree(call_state);
	return;
}

void pla_finish_verify_continue_sign(void *call_state)
{
	struct pla_call_state *state = (struct pla_call_state*) call_state;

	/* Complete the verification (window, timestamp, etc...) */
	if (complete_verify(call_state) < 0 || pla_add_header(&state->skb) < 0)
		goto out_drop;

	/* Update pointer to new header */
	state->pla_hdr = pla_find_pla_header(state->skb);

	if (pla_setup_crypto_call_sign(state->skb, state->info, state->pla_hdr, pla_out_path_finish) < 0)
		goto out_drop;

	kfree(call_state);
	return;

out_drop:
	pla_immediate_verdict_state(state, NF_DROP);
	kfree(call_state);
	return;
}

int pla_verify_and_add_header(struct sk_buff **skb, struct pla_hdr *pla_hdr, struct nf_info *info)
{
	int retval = 0;

	if (pla_setup_crypto_call_verify(*skb, info, pla_hdr, CRYPTO_TASK_TYPE_VERIFY, &pla_finish_verify_continue_sign) < 0) {
		retval = -1;
		goto out;
	}


out:
	return retval;
}

/*
 * Return values:
 *  1:    accept immediately
 *  0:    continue with crypto path
 * <= -1: drop packet immediately
 */
signed int pla_forward_path(struct sk_buff **skb, struct nf_info *info, struct pla_hdr *pla_hdr)
{
	int err = -1;
	int policy;

	policy = pla_forward_policy(info);

	if (pla_hdr) {
		switch(policy) {
		case PLA_FORWARD_POLICY_NONE:    /* No policy set: Try to verify */
		case PLA_FORWARD_POLICY_VERIFY:
			if (pla_setup_crypto_call_verify(*skb, info, pla_hdr, CRYPTO_TASK_TYPE_VERIFY, &pla_finish_verify) < 0)
				goto drop_out;
			break;

		case PLA_FORWARD_POLICY_STRIP:
			/* Verify packet before stripping */
			if (pla_setup_crypto_call_verify(*skb, info, pla_hdr, CRYPTO_TASK_TYPE_VERIFY, &pla_finish_verify_strip) < 0)
				goto drop_out;

#if 0
			err = pla_strip_pla_header(skb);
			if (err < 0)
				goto drop_out;
#endif
			break;

		case PLA_FORWARD_POLICY_SIGN:
			/* Packet already has PLA. Verify the existing header and add new header. */
			if (pla_verify_and_add_header(skb, pla_hdr, info) < 0)
				goto drop_out;
			break;

		case PLA_FORWARD_POLICY_DROP:
			goto drop_out;
			break;

		default:
			/* Can't happen */
			BUG();
			break;
		}

	} else {
		switch (policy) {
		case PLA_FORWARD_POLICY_NONE:    /* No policy set: Add PLA */
		case PLA_FORWARD_POLICY_SIGN:
			if (pla_add_header(skb) < 0) 
				goto drop_out;
			pla_hdr = pla_find_pla_header(*skb);
			if ( (err = pla_setup_crypto_call_sign(*skb, info, pla_hdr, pla_out_path_finish)) < 0)
				goto drop_out;
			break;

		case PLA_FORWARD_POLICY_STRIP:
			/* No PLA header in packet. Accept. */
			goto accept_out;
			break;

		case PLA_FORWARD_POLICY_DROP:
		case PLA_FORWARD_POLICY_VERIFY:
			/* PLA verification requuired, but no PLA in packet */
			goto drop_out;
			break;
			
		default:
			/* Can't happen. */
			BUG();
			break;
		}

	}

	return 0;

accept_out:
	err = 1;
	return err;

drop_out:

	return err;
}

int pla_in_path(struct sk_buff *skb, struct nf_info *info, struct pla_hdr *pla_hdr)
{
	int err;

	/* 
	 * Netfilter stupidity: We cannot get only those packets which have a nexthdr field
	 * of Hop-By-Hop, because its value is 0 which means "all" to netfilter.
	 */

	/* Drop packets that were given to PLA if the packet has no PLA header. */
	if (!pla_hdr) {
		err = -1;
		goto out_err;
	}

	if ( (err = pla_setup_crypto_call_verify(skb, info, pla_hdr, CRYPTO_TASK_TYPE_VERIFY, &pla_finish_verify)) < 0) {
		err = -1;
		goto out_err;
	}

	return 0;

out_err:
	return err;
}


int pla_queue_enqueue(struct sk_buff *skb, struct nf_info *info, unsigned int queuenum, void *data)
{
	signed int err = 0;
	struct pla_hdr *pla_hdr;

	atomic_inc(&pla_stats_total_packets);
	atomic_inc(&processing_packets);
	
	/* No PLA on the loopback device */
	if (info->indev == &loopback_dev || info->outdev == &loopback_dev) {
		pla_immediate_verdict(skb, info, NF_ACCEPT);
		return 0;
	}

#if 1
	/* Don't let queues become too long */
	if (atomic_read(&processing_packets) >= 50) {
		pla_immediate_verdict(skb, info, NF_DROP);
		return 0;
	}
#endif

	pla_hdr = pla_find_pla_header(skb);

	switch(info->hook) {
	case NF_IP6_LOCAL_OUT:
		err = pla_out_path(&skb, info, pla_hdr);
		break;

	case NF_IP6_FORWARD:
		err = pla_forward_path(&skb, info, pla_hdr);
		break;

	case NF_IP6_LOCAL_IN:
		err = pla_in_path(skb, info, pla_hdr);
		break;

	default:
		err = -1;
		break;
	}

	if (err > 0)
		pla_immediate_verdict(skb, info, NF_ACCEPT);

	if (err < 0)
		pla_immediate_verdict(skb, info, NF_DROP);

	return 0;
}

struct nf_queue_handler pla_queue = {
        .outfn = &pla_queue_enqueue,
        .name = "pla_queue"
};

static int init_or_exit(int is_init)
{
	int err = 0;
	wait_queue_head_t wqh;

	if (!is_init)
		goto cleanup_exit;

	err = pla_state_init();
	if (err < 0)
		goto state_init_err;

        /* None of these can fail */
        pla_src_id_init();
        pla_policy_init();
	pla_conn_init();

	if(!create_proc_read_entry(PLA_PROC_STAT_NAME, 0, NULL, pla_read_stats_proc, NULL)) {
		err = -ENOMEM;
		goto create_proc_failed;
	}

	err = pla_crypto_init(CRYPTO_IMPL_USERSPACE);
	if (err)
		goto init_crypto_failed;

	err = nf_register_queue_handler(AF_INET6, &pla_queue);
	if (err)
		goto register_queue_failed;

	if (UNREGISTER_REROUTER() < 0)
		goto unregister_rerouter_failed;

	err = pla_netlink_init();
	if (err)
		goto netlink_init_failed;

	return 0;

cleanup_exit:

	/* Nothing to do here... */
netlink_init_failed:
	pla_netlink_uninit();

unregister_rerouter_failed:
	nf_unregister_queue_handler(AF_INET6);

	/* Wait until all packets complete processing */
	init_waitqueue_head(&wqh);
	while (atomic_read(&processing_packets) != 0)
		wait_event_timeout(wqh, 0, HZ/10);
register_queue_failed:

	pla_crypto_uninit();
init_crypto_failed:

	remove_proc_entry(PLA_PROC_STAT_NAME, NULL);
create_proc_failed:

	pla_conn_uninit();
        pla_policy_uninit();
        pla_src_id_uninit();

	pla_state_uninit();
state_init_err:

	return err;
}

static int pla_init(void)
{
	return init_or_exit(1);
}

static void pla_exit(void)
{
	init_or_exit(0);
	return;
}

MODULE_AUTHOR("Janne Lundberg <jlu@tcs.hut.fi>");
MODULE_DESCRIPTION("IPv6 Packet Level Authentication.");
MODULE_LICENSE("GPL");

module_init(pla_init);
module_exit(pla_exit);
