#include <linux/netlink.h>
#include <linux/netdevice.h>
#include <linux/wait.h>
#include <linux/spinlock.h>

#include "pla_src_id.h"
#include "pla_netlink.h"
#include "pla_state.h"

#include "configs.h"

struct pla_src_id 
{
	/* Mandatory housekeeping */
	struct list_head list;        /* Must be first in struct */
	struct rcu_head rcu_head;

	/* Payload */
	int pub_len;                  /* Length of the public key  */
	int priv_len;                 /* Length of the private key */
	int cert_len;                 /* Length of the certificate */
	int id_hash_len;

	/* 
	 * Pointers are to memory allocated with the same kmalloc as this struct.
	 * That is, their memory is freed when this struct is freed.
	 */
	char *pub_ptr;                /* Pointer to public key. */
	char *priv_ptr;               /* Pointer to private key */
	char *cert_ptr;               /* Pointer to certificate for the key */
	char *id_hash_ptr;

	uint64_t    seqnum;           
	spinlock_t  seqnum_lock;
};

LIST_HEAD(src_id);
spinlock_t src_id_write_lock;
static atomic_t uncleaned_src_entries = ATOMIC_INIT(0);  /* Number of uncleaned entries in src_id list */

static void pla_src_id_free_entry(struct rcu_head *head)
{
	kfree(container_of(head, struct pla_src_id, rcu_head));
	atomic_dec(&uncleaned_src_entries);

	return;
}

static int pla_src_id_flush_complete(void)
{
	return atomic_read(&uncleaned_src_entries) == 0 ? 1 : 0;
}

static void pla_src_id_flush(void)
{
	struct pla_src_id *e;
	struct pla_src_id *safe_storage;

	list_for_each_entry_safe(e, safe_storage, &src_id, list) {
		list_del_rcu(&e->list);
		call_rcu( &e->rcu_head, pla_src_id_free_entry);
	}

	return;
}

static struct pla_src_id* get_first_entry(void) 
{
	struct list_head *pos;
	
	/* jlu XXX: Yes, list_for_each_rcu makes no sense here. TBD */
	list_for_each_rcu(pos, &src_id)
		return (struct pla_src_id*) pos;

	return NULL; /* If no entries in list */
}

int pla_src_key_copy(char *pubkey, char *privkey, char *cert, uint64_t *seqnum, char *id_hash, uint32_t *ttp_id) 
{
	int ret = 0;
	struct pla_src_id *entry;

	rcu_read_lock();
	
	entry = get_first_entry();
	if (!entry) {
		ret = 1;
		goto out;
	}

	if (pubkey)
		memcpy(pubkey, entry->pub_ptr, PLA_PUBKEY_LEN);

	if (privkey)
		memcpy(privkey, entry->priv_ptr, PLA_PRIVKEY_LEN);

	if (cert)
		memcpy(cert, entry->cert_ptr, PLA_CERT_LEN);

	if (seqnum) {
		spin_lock_bh(&entry->seqnum_lock);
		*seqnum = cpu_to_be64(entry->seqnum);
		entry->seqnum++;
		spin_unlock_bh(&entry->seqnum_lock);
	}
	
	if (id_hash)
		memcpy(id_hash, entry->id_hash_ptr, PLA_HASH_DIGEST_LENGTH);

	if (ttp_id) 
		*ttp_id = htonl(0x12345678);

out:
	rcu_read_unlock();

	return ret;
}

static int pla_src_id_add(char *pubkey, int publen, char *privkey, int privlen, char *certificate, int certlen,
			  char *id_hash, int id_hashlen)
{
	int err = 0;
	struct pla_src_id *entry;

	entry = kmalloc(sizeof(struct pla_src_id) + publen + privlen + certlen + id_hashlen, GFP_ATOMIC);
	if (!entry) {
		err = -ENOMEM;
		goto out;
	}
	
	entry->pub_len = publen;
	entry->pub_ptr = (char*) (entry + 1);
	memcpy(entry->pub_ptr, pubkey, publen);

	entry->priv_len = privlen;
	entry->priv_ptr = entry->pub_ptr + publen;
	memcpy(entry->priv_ptr, privkey, privlen);

	entry->cert_len = certlen;
	entry->cert_ptr = entry->priv_ptr + privlen;
	memcpy(entry->cert_ptr, certificate, certlen);

	entry->id_hash_len = id_hashlen;
	entry->id_hash_ptr = entry->cert_ptr + certlen;
	memcpy(entry->id_hash_ptr, id_hash, id_hashlen);
	
	entry->seqnum = 0;
	spin_lock_init(&entry->seqnum_lock);

	spin_lock_bh(&src_id_write_lock);
	list_add_rcu( &entry->list, &src_id);
	atomic_inc(&uncleaned_src_entries);
	spin_unlock_bh(&src_id_write_lock);

out:
	return err;
}

int pla_src_id_netlink(struct nlmsghdr *nlh)
{
	int err = 0;
	int type;
	struct pla_netlink_add_src *new_src;

	char *pubkey;
	char *privkey;
	char *cert;
	char *id_hash;

	type = nlh->nlmsg_type;
	switch(type) {
	case NLMSG_PLA_SET_SRC:
		new_src = NLMSG_DATA(nlh);

		pubkey = (char*) (new_src + 1);
		privkey = pubkey + new_src->publen;
		cert = privkey + new_src->privlen;
		id_hash = cert + new_src->certlen;

		// jlu XXX: What happens on error? Is this the right place for this call?
		pla_state_add(pubkey, new_src->publen);

		pla_src_id_add(pubkey, new_src->publen, 
			       privkey, new_src->privlen, 
			       cert, new_src->certlen,
			       id_hash, new_src->id_hashlen);
		break;

	default:
		break;
	}

	return err;
}

void pla_src_id_uninit(void)
{
	wait_queue_head_t wqh;
	
	pla_src_id_flush();
	/* 
	 * Wait until all RCU-callbacks have completed.
	 * If the module exits too early, the callbacks will call nonexisting functions.
	 */
	init_waitqueue_head(&wqh);
	while (!pla_src_id_flush_complete())
		wait_event_timeout(wqh, 0, HZ/10);
	
	return;
}

int pla_src_id_init(void)
{
	spin_lock_init(&src_id_write_lock);

	return 0;
}
